#ifndef ABSTRACT_H
#define ABSTRACT_H

#define HostOS 0x01

// Memory abstraction functions

bool _RamLoad(char* filename, uint16 address)
{
	File f;
	bool result = false;

	if (f = SD.open(filename, FILE_READ)) {
		while (f.available()) {
			_RamWrite(address++, f.read());
		}
		f.close();
		result = true;
	}
	return result;
}


void _RamFill(uint16 address, int size, uint8 value)
{
	while (size--) {
		_RamWrite(address++, value);
	}
}


void _RamCopy(uint16 source, int size, uint16 destination)
{
	while (size--) {
		_RamWrite(destination++, _RamRead(source++));
	}
}


// Filesystem (disk) abstraction fuctions

File rootdir, userdir;
#define FOLDERCHAR '/'

typedef struct {
	uint8 dr;
	uint8 fn[8];
	uint8 tp[3];
	uint8 ex, s1, s2, rc;
	uint8 al[16];
	uint8 cr, r0, r1, r2;
} CPM_FCB;

typedef struct {
	uint8 dr;
	uint8 fn[8];
	uint8 tp[3];
	uint8 ex, s1, s2, rc;
	uint8 al[16];
} CPM_DIRENTRY;

static dir_t fileDirEntry;


File _sys_fopen_w(uint8* filename)
{
	return SD.open((char*)filename, O_CREAT | O_WRITE);
}


int _sys_fputc(uint8 ch, File& f)
{
	return f.write(ch);
}


void _sys_fflush(File& f)
{
	f.flush();
}


void _sys_fclose(File& f)
{
	f.close();
}


int _sys_select(uint8* disk)
{
	uint8 result = false;
	File f;

	digitalWrite(LED, HIGH);
	if (f = SD.open((char*)disk, O_READ)) {
		if (f.isDirectory()) {
			result = true;
		}
		f.close();
	}
	digitalWrite(LED, LOW);
	return result;
}


long _sys_filesize(uint8* filename)
{
	long l = -1;
	File f;

	digitalWrite(LED, HIGH);
	if (f = SD.open((char*)filename, O_RDONLY)) {
		l = f.size();
		f.close();
	}
	digitalWrite(LED, LOW);
	return l;
}


int _sys_openfile(uint8* filename)
{
	File f;
	int result = 0;

	digitalWrite(LED, HIGH);
	f = SD.open((char*)filename, O_READ);
	if (f) {
		f.dirEntry(&fileDirEntry);
		f.close();
		result = 1;
	}
	digitalWrite(LED, LOW);
	return result;
}


int _sys_makefile(uint8* filename)
{
	File f;
	int result = 0;

	digitalWrite(LED, HIGH);
	f = SD.open((char*)filename, O_CREAT | O_WRITE);
	if (f) {
		f.close();
		result = 1;
	}
	digitalWrite(LED, LOW);
	return result;
}


int _sys_deletefile(uint8* filename)
{
	digitalWrite(LED, HIGH);
	return SD.remove((char*)filename);
	digitalWrite(LED, LOW);
}


int _sys_renamefile(uint8* filename, uint8* newname)
{
	File f;
	int result = 0;

	digitalWrite(LED, HIGH);
	f = SD.open((char*)filename, O_WRITE | O_APPEND);
	if (f) {
    if (f.rename(SD.vwd(), (char*)newname)) {
			f.close();
			result = 1;
		}
	}
	digitalWrite(LED, LOW);
	return result;
}


bool _sys_extendfile(char* fn, unsigned long fpos)
{
	uint8 result = true;
	File f;
	unsigned long i;

	digitalWrite(LED, HIGH);
	if (f = SD.open(fn, O_WRITE | O_APPEND)) {
		if (fpos > f.size()) {
			for (i = 0; i < f.size() - fpos; ++i) {
				if (f.write((uint8)0) != 1) {
					result = false;
					break;
				}
			}
		}
		f.close();
	} else {
		result = false;
	}
	digitalWrite(LED, LOW);
	return result;
}


uint8 _sys_readseq(uint8* filename, long fpos)
{
	uint8 result = 0xff;
	File f;
	uint8 bytesread;
	uint8 dmabuf[BlkSZ];
	uint8 i;

	digitalWrite(LED, HIGH);
	f = SD.open((char*)filename, O_READ);
	if (f) {
		if (f.seek(fpos)) {
			for (i = 0; i < BlkSZ; ++i)
				dmabuf[i] = 0x1a;
			bytesread = f.read(&dmabuf[0], BlkSZ);
			if (bytesread) {
				for (i = 0; i < BlkSZ; ++i)
					_RamWrite(dmaAddr + i, dmabuf[i]);
			}
			result = bytesread ? 0x00 : 0x01;
		} else {
			result = 0x01;
		}
		f.close();
	} else {
		result = 0x10;
	}
	digitalWrite(LED, LOW);
	return result;
}


uint8 _sys_writeseq(uint8* filename, long fpos)
{
	uint8 result = 0xff;
	File f;

	digitalWrite(LED, HIGH);
	if (_sys_extendfile((char*)filename, fpos))
		f = SD.open((char*)filename, O_RDWR);
	if (f) {
		if (f.seek(fpos)) {
			if (f.write(_RamSysAddr(dmaAddr), BlkSZ)) {
				result = 0x00;
			}
		} else {
			result = 0x01;
		}
		f.close();
	} else {
		result = 0x10;
	}
	digitalWrite(LED, LOW);
	return result;
}


uint8 _sys_readrand(uint8* filename, long fpos)
{
	uint8 result = 0xff;
	File f;
	uint8 bytesread;
	uint8 dmabuf[BlkSZ];
	uint8 i;
	long extSize;

	digitalWrite(LED, HIGH);
	f = SD.open((char*)filename, O_READ);
	if (f) {
		if (f.seek(fpos)) {
			for (i = 0; i < BlkSZ; ++i) {
				dmabuf[i] = 0x1a;
			}
			bytesread = f.read(&dmabuf[0], BlkSZ);
			if (bytesread) {
				for (i = 0; i < BlkSZ; ++i) {
					_RamWrite(dmaAddr + i, dmabuf[i]);
				}
			}
			result = bytesread ? 0x00 : 0x01;
		} else {
			if (fpos >= 65536L * BlkSZ) {
				result = 0x06;	// seek past 8MB (largest file size in CP/M)
			} else {
				extSize = f.size();
				// round file size up to next full logical extent
				extSize = ExtSZ * ((extSize / ExtSZ) + ((extSize % ExtSZ) ? 1 : 0));
				if (fpos < extSize) {
					result = 0x01;	// reading unwritten data
				} else {
					result = 0x04; // seek to unwritten extent
				}
			}
		}
		f.close();
	} else {
		result = 0x10;
	}
	digitalWrite(LED, LOW);
	return result;
}


uint8 _sys_writerand(uint8* filename, long fpos)
{
	uint8 result = 0xff;
	File f;

	digitalWrite(LED, HIGH);
	if (_sys_extendfile((char*)filename, fpos)) {
		f = SD.open((char*)filename, O_RDWR);
	}
	if (f) {
		if (f.seek(fpos)) {
			if (f.write(_RamSysAddr(dmaAddr), BlkSZ)) {
				result = 0x00;
			}
		} else {
			result = 0x06;
		}
		f.close();
	} else {
		result = 0x10;
	}
	digitalWrite(LED, LOW);
	return result;
}


static uint8 findNextDirName[13];
static uint16 fileRecords = 0;
static uint16 fileExtents = 0;
static uint16 fileExtentsUsed = 0;
static uint16 firstFreeAllocBlock;

uint8 _findnext(uint8 isdir)
{
	File f;
	uint8 result = 0xff;
	bool isfile;
	uint32 bytes;

	digitalWrite(LED, HIGH);
	if (allExtents && fileRecords) {
		_mockupDirEntry();
		result = 0;
	} else {
		while (f = userdir.openNextFile()) {
			f.getName((char*)&findNextDirName[0], 13);
			isfile = !f.isDirectory();
			bytes = f.size();
			f.dirEntry(&fileDirEntry);
			f.close();
			if (!isfile)
				continue;
			_HostnameToFCBname(findNextDirName, fcbname);
			if (match(fcbname, pattern)) {
				if (isdir) {
					// account for host files that aren't multiples of the block size
					// by rounding their bytes up to the next multiple of blocks
					if (bytes & (BlkSZ - 1)) {
						bytes = (bytes & ~(BlkSZ - 1)) + BlkSZ;
					}
					fileRecords = bytes / BlkSZ;
					fileExtents = fileRecords / BlkEX + ((fileRecords & (BlkEX - 1)) ? 1 : 0);
					fileExtentsUsed = 0;
					firstFreeAllocBlock = firstBlockAfterDir;
					_mockupDirEntry();
				} else {
					fileRecords = 0;
					fileExtents = 0;
					fileExtentsUsed = 0;
					firstFreeAllocBlock = firstBlockAfterDir;
				}
				_RamWrite(tmpFCB, filename[0] - '@');
				_HostnameToFCB(tmpFCB, findNextDirName);
				result = 0x00;
				break;
			}
		}
	}
	digitalWrite(LED, LOW);
	return result;
}


uint8 _findfirst(uint8 isdir)
{
	uint8 path[4] = { '?', FOLDERCHAR, '?', 0 };
	path[0] = filename[0];
	path[2] = filename[2];
	if (userdir) {
		userdir.close();
	}
	userdir = SD.open((char*)path); // Set directory search to start from the first position
	_HostnameToFCBname(filename, pattern);
	fileRecords = 0;
	fileExtents = 0;
	fileExtentsUsed = 0;
	return _findnext(isdir);
}


uint8 _findnextallusers(uint8 isdir)
{
	uint8 result = 0xFF;
	char dirname[13];
	bool done = false;

	while (!done) {
		while (!userdir) {
			userdir = rootdir.openNextFile();
			if (!userdir) {
				done = true;
				break;
			}
			userdir.getName(dirname, sizeof dirname);
			if (userdir.isDirectory() && strlen(dirname) == 1 && isxdigit(dirname[0])) {
				currFindUser = dirname[0] <= '9' ? dirname[0] - '0' : toupper(dirname[0]) - 'A' + 10;
				break;
			}
			userdir.close();
		}
		if (userdir) {
			result = _findnext(isdir);
			if (result) {
				userdir.close();
			} else {
				done = true;
			}
		} else {
			result = 0xFF;
			done = true;
		}
	}
	return result;
}


uint8 _findfirstallusers(uint8 isdir)
{
	uint8 path[2] = { '?', 0 };

	path[0] = filename[0];
	if (rootdir)
		rootdir.close();
	if (userdir)
		userdir.close();
	rootdir = SD.open((char*)path); // Set directory search to start from the first position
	_HostnameToFCBname(filename, pattern); // ##### changed
	if (!rootdir) {
		return 0xff;
	}
	fileRecords = 0;
	fileExtents = 0;
	fileExtentsUsed = 0;
	return _findnextallusers(isdir);
}


uint8 _Truncate(char* filename, uint8 rc)
{
	File f;
	int result = 0;

	digitalWrite(LED, HIGH);
	f = SD.open((char*)filename, O_WRITE | O_APPEND);
	if (f) {
		if (f.truncate(rc * BlkSZ)) {
			f.close();
			result = 1;
		}
	}
	digitalWrite(LED, LOW);
	return result;
}


void _MakeUserDir()
{
	uint8 dFolder = cDrive + 'A';
	uint8 uFolder = toupper(tohex(userCode));

	uint8 path[4] = { dFolder, FOLDERCHAR, uFolder, 0 };

	digitalWrite(LED, HIGH);
	SD.mkdir((char*)path);
	digitalWrite(LED, LOW);
}


uint8 _sys_makedisk(uint8 drive)
{
	uint8 result = 0;
	if (drive < 1 || drive > 16) {
		result = 0xff;
	} else {
		uint8 dFolder = drive + '@';
		uint8 disk[2] = { dFolder, 0 };
		digitalWrite(LED, HIGH);
		if (!SD.mkdir((char*)disk)) {
			result = 0xfe;
		} else {
			uint8 path[4] = { dFolder, FOLDERCHAR, '0', 0 };
			SD.mkdir((char*)path);
		}
		digitalWrite(LED, LOW);
	}
	return result;
}


// Console abstraction functions ##### changed and added

int _kbhit()
{
	if (_RamRead(IObyte) & 0x03) {
		return Serial.available();
	} else {
		return Serial1.available();
	}
}


uint8 _getch()
{
	if (_RamRead(IObyte) & 0x03) {
		while (!Serial.available()) ;
		return Serial.read();
	} else {
		while (!Serial1.available()) ;
		return Serial1.read();
	}
}


void _putch(uint8 ch)
{
	if (_RamRead(IObyte) & 0x03) {
		Serial.write(ch);
	} else {
		Serial1.write(ch);
	}
}


void _putaux(uint8 ch)
{
	Serial1.write(ch);
}


#define XON  0x11

// output with Xon/Xoff protocol
void _putauxxon(uint8 ch)
{
	bool portBusy = Serial1.available();
	while (portBusy) {
		while (!Serial1.available()) ;
		portBusy = (Serial1.read() != XON);
	}
	Serial1.write(ch);
}


uint8 _getaux()
{
	while (!Serial1.available()) ;
	return Serial1.read();
}


void hw_Reset()
{
	pinMode(RESETPIN, OUTPUT);
	digitalWrite(RESETPIN, LOW);
}


// parallel port pin# = arduino pin#
#define nStrobe 23
#define data_0 25
#define data_1 27
#define data_2 29
#define data_3 31
#define data_4 33
#define data_5 35
#define data_6 37
#define data_7 39
// #define nAck 41
#define busy 43

// parallel port init
void _initlst()
{
	pinMode(nStrobe, OUTPUT);      // is active LOW
	digitalWrite(nStrobe, HIGH);
	pinMode(data_0, OUTPUT);
	pinMode(data_1, OUTPUT);
	pinMode(data_2, OUTPUT);
	pinMode(data_3, OUTPUT);
	pinMode(data_4, OUTPUT);
	pinMode(data_5, OUTPUT);
	pinMode(data_6, OUTPUT);
	pinMode(data_7, OUTPUT);
	// pinMode(nAck, INPUT);     // is active LOW
	pinMode(busy, INPUT);
}


// parallel port output
void _putlst(uint8 ch)
{
	while (digitalRead(busy) == HIGH) ;  // wait for busy to go low

	digitalWrite(data_0, ch & 0x01);
	digitalWrite(data_1, ch & 0x02);
	digitalWrite(data_2, ch & 0x04);
	digitalWrite(data_3, ch & 0x08);
	digitalWrite(data_4, ch & 0x10);
	digitalWrite(data_5, ch & 0x20);
	digitalWrite(data_6, ch & 0x40);
	digitalWrite(data_7, ch & 0x80);
	digitalWrite(nStrobe, LOW);         // strobe nStrobe to input data bits
	delayMicroseconds(2);
	digitalWrite(nStrobe, HIGH);
}


void halt()
{
	_puts("CPU halted.\r\n");
	while (true) {
		digitalWrite(LED, HIGH);
		delay(DELAY);
		digitalWrite(LED, LOW);
		delay(DELAY);
		digitalWrite(LED, HIGH);
		delay(DELAY);
		digitalWrite(LED, LOW);
		delay(DELAY * 4);
	}
}

#endif
