// RunCPM extensions
// printing using ANSI print commands

#ifndef ANSIPRINT_H
#define ANSIPRINT_H

bool printJob = false;
unsigned long lastChar;
#define BUFSIZE 2048
uint8 conBuf[BUFSIZE];
uint16 bufIdx = 0;
uint16 bI;

// Put null terminated string
void putStr(const char* str)
{
	while (*str) {
		_putch(*(str++));
	}
}

// Insert char for console output into buffer
void insBuf(uint8 ch)
{
	if (bufIdx < BUFSIZE) {
		conBuf[bufIdx++] = ch;
	}
}

// Timer handler for sending chars from buffer
void bufHandler()
{
	if (bI == bufIdx) {
		Timer6.stop();     // stop timer
		printJob = false;  // reset print job flag
		bufIdx = 0;
	} else {
		_putch(conBuf[bI++]);
	}
}

// Timer handler for checking time of the last char
void printHandler()
{
	// more than 500 ms past the last char output -> close print job
	if ((millis() - lastChar) > 500) {
		Timer6.stop();     // stop timer
		putStr("\e[4i");   // ANSI string for end of stream capture
		bI = 0;            // prepare buffer chars sending
		Timer6.attachInterrupt(bufHandler).start(1200);
	}
}

// "Printing" routine
void ansiPrint(uint8 ch)
{
	// start print job
	if (!printJob) {
		printJob = true;   // set print job flag
		putStr("\e[5i");   // ANSI string for start of stream capture
		Timer6.attachInterrupt(printHandler).start(100000); // calls every 100 ms
	}
	_putch(ch);          // char output
	lastChar = millis(); // mark output time
}

#endif
