#ifndef CPM_H
#define CPM_H

#define JP		0xc3
#define CALL	0xcd
#define RET		0xc9
#define INa		0xdb	// Triggers a BIOS call
#define OUTa	0xd3	// Triggers a BDOS call

unsigned long freq = 4000; //##### added  default value

void _PatchCPM() {
	uint16 i;

	//**********  Patch CP/M page zero into the memory  **********
	_RamWrite(0x0000, JP);		/* JP BIOS+3 (warm boot) */
	_RamWrite16(0x0001, BIOSjmppage + 3);
	if (Status != 2) {
		_RamWrite(DUbyte, 0x00);
	}
	_RamWrite(0x0005, JP);
	_RamWrite16(0x0006, BDOSpage + 0x06);
	_RamLoad((char *)"A/0/BDOS.SYS", BDOSpage);

	//**********  Patch CP/M (fake) Disk Paramater Table after the BDOS call entry  **********
	i = DPBaddr;
	_RamWrite(i++, 64);  		/* spt - Sectors Per Track */
	_RamWrite(i++, 0);
	_RamWrite(i++, 5);   		/* bsh - Data allocation "Block Shift Factor" */
	_RamWrite(i++, 0x1F);		/* blm - Data allocation Block Mask */
	_RamWrite(i++, 1);   		/* exm - Extent Mask */
	_RamWrite(i++, 0xFF);		/* dsm - Total storage capacity of the disk drive */
	_RamWrite(i++, 0x07);
	_RamWrite(i++, 255); 		/* drm - Number of the last directory entry */
	_RamWrite(i++, 3);
	_RamWrite(i++, 0xFF);		/* al0 */
	_RamWrite(i++, 0x00);		/* al1 */
	_RamWrite(i++, 0);   		/* cks - Check area Size */
	_RamWrite(i++, 0);
	_RamWrite(i++, 0x02);		/* off - Number of system reserved tracks at the beginning of the ( logical ) disk */
	_RamWrite(i++, 0x00);
	blockShift = _RamRead(DPBaddr + 2);
	blockMask = _RamRead(DPBaddr + 3);
	extentMask = _RamRead(DPBaddr + 4);
	numAllocBlocks = _RamRead16((DPBaddr + 5)) + 1;
	extentsPerDirEntry = extentMask + 1;

	// figure out the number of the first allocation block
	// after the directory for the phoney allocation block
	// list in _findnext()
	firstBlockAfterDir = 0;
	i = 0x80;
	while (_RamRead(DPBaddr + 9) & i) {
		firstBlockAfterDir++;
		i >>= 1;
	}
	if (_RamRead(DPBaddr + 9) == 0xFF) {
		i = 0x80;
		while (_RamRead(DPBaddr + 10) & i) {
			firstBlockAfterDir++;
			i >>= 1;
		}
	}
	physicalExtentBytes = logicalExtentBytes * (extentMask + 1);
}


void _PatchBIOS()
{
	uint16 i;

	for (i = 0; i < 0x36; i = i + 3) {
		_RamWrite(BIOSjmppage + i, JP);
		_RamWrite16(BIOSjmppage + i + 1, BIOSpage + i);
		_RamWrite(BIOSpage + i, OUTa);
		_RamWrite(BIOSpage + i + 1, i & 0xff);
		_RamWrite(BIOSpage + i + 2, RET);
	}
}


void _Bios()
{
	uint8 ch = LOW_REGISTER(BC);

	switch (LOW_REGISTER(PCX)) {
	case 0x00:
		Status = 1;			// 0 - BOOT - Ends RunCPM
		break;
	case 0x03:
		Status = 2;			// 1 - WBOOT - Back to CCP
		break;
	case 0x06:				// 2 - CONST - Console status
		SET_HIGH_REGISTER(AF, _kbhit() ? 0xff : 0x00);
		break;
	case 0x09:				// 3 - CONIN - Console input
		SET_HIGH_REGISTER(AF, _getch());
		break;
	case 0x0C:				// 4 - CONOUT - Console output
		_putcon(ch);
		break;
	case 0x0F:				// 5 - LIST - List output
		// ##### changed
		switch (_RamRead(IObyte) & 0xc0) {
			case 0x00:
				_putauxxon(ch);
				break;
			case 0x40:
				_putcon(ch);
				break;
			case 0x80:
				_putlst(ch);
				break;
			case 0xc0:
				ansiPrint(ch);
				break;
		}
		// ##### end
		break;
	case 0x12:				// 6 - PUNCH/AUXOUT - Punch output
		// ##### changed
		if (!(_RamRead(IObyte) & 0x30)) {
			_putaux(ch);
		}
		// ##### end
		break;
	case 0x15:				// 7 - READER - Reader input (0x1a = device not implemented)
		// ##### changed
		if (_RamRead(IObyte) & 0x0c) {
			SET_HIGH_REGISTER(AF, 0x1a);
		} else {
			SET_HIGH_REGISTER(AF, _getaux());
		// ##### end
		}
		break;
	case 0x18:				// 8 - HOME - Home disk head
		break;
	case 0x1B:				// 9 - SELDSK - Select disk drive
		HL = 0x0000;
		break;
	case 0x1E:				// 10 - SETTRK - Set track number
		break;
	case 0x21:				// 11 - SETSEC - Set sector number
		break;
	case 0x24:				// 12 - SETDMA - Set DMA address
		HL = BC;
		dmaAddr = BC;
		break;
	case 0x27:				// 13 - READ - Read selected sector
		SET_HIGH_REGISTER(AF, 0x00);
		break;
	case 0x2A:				// 14 - WRITE - Write selected sector
		SET_HIGH_REGISTER(AF, 0x00);
		break;
	case 0x2D:				// 15 - LISTST - Get list device status
		SET_HIGH_REGISTER(AF, 0xff);
		// ##### changed
		if (((_RamRead(IObyte) & 0xc0) == 0x80) & getListSt()) {
			SET_HIGH_REGISTER(AF, 0x00);
		}
		// ##### end
		break;
	case 0x30:				// 16 - SECTRAN - Sector translate
		HL = BC;				// HL=BC=No translation (1:1)
		break;
	case 0x33:				// 17 - RETTOCCP - This allows programs ending in RET return to internal CCP
		Status = 3;
		break;
	default:
		break;
	}
}


void _Bdos() {
	HL = 0x0000;	// HL is reset by the BDOS

	switch (LOW_REGISTER(BC)) {
	  /*
		C = 13 (0Dh) : Reset disk system
		*/
	case 13:
		roVector = 0;		// Make all drives R/W
		loginVector = 0;
		dmaAddr = 0x0080;
		cDrive = 0;			// userCode remains unchanged
		HL = _CheckSUB();	// Checks if there's a $$$.SUB on the boot disk
		break;
		/*
		C = 14 (0Eh) : Select Disk
		Returns: A=0x00 or 0xFF
		*/
	case 14:
		oDrive = cDrive;
		cDrive = LOW_REGISTER(DE);
		HL = _SelectDisk(LOW_REGISTER(DE) + 1);	// +1 here is to allow SelectDisk to be used directly by disk.h as well
		if (!HL)
			oDrive = cDrive;
		break;
		/*
		C = 15 (0Fh) : Open file
		Returns: A=0x00 or 0xFF
		*/
	case 15:
		HL = _OpenFile(DE);
		break;
		/*
		C = 16 (10h) : Close file
		*/
	case 16:
		HL = _CloseFile(DE);
		break;
		/*
		C = 17 (11h) : Search for first
		*/
	case 17:
		HL = _SearchFirst(DE, true);	// TRUE = Creates a fake dir entry when finding the file
		break;
		/*
		C = 18 (12h) : Search for next
		*/
	case 18:
		HL = _SearchNext(DE, true);		// TRUE = Creates a fake dir entry when finding the file
		break;
		/*
		C = 19 (13h) : Delete file
		*/
	case 19:
		HL = _DeleteFile(DE);
		break;
		/*
		C = 20 (14h) : Read sequential
		*/
	case 20:
		HL = _ReadSeq(DE);
		break;
		/*
		C = 21 (15h) : Write sequential
		*/
	case 21:
		HL = _WriteSeq(DE);
		break;
		/*
		C = 22 (16h) : Make file
		*/
	case 22:
		HL = _MakeFile(DE);
		break;
		/*
		C = 23 (17h) : Rename file
		*/
	case 23:
		HL = _RenameFile(DE);
		break;
		/*
		C = 24 (18h) : Return log-in vector (active drive map)
		*/
	case 24:
		HL = loginVector;	// (todo) improve this
		break;
		/*
		C = 25 (19h) : Return current disk
		*/
	case 25:
		HL = cDrive;
		break;
		/*
		C = 26 (1Ah) : Set DMA address
		*/
	case 26:
		dmaAddr = DE;
		break;
		/*
		C = 27 (1Bh) : Get ADDR(Alloc)
		*/
	case 27:
		HL = SCBaddr;
		break;
		/*
		C = 28 (1Ch) : Write protect current disk
		*/
	case 28:
		roVector = roVector | (1 << cDrive);
		break;
		/*
		C = 29 (1Dh) : Get R/O vector
		*/
	case 29:
		HL = roVector;
		break;
		/*
		C = 30 (1Eh) : Set file attributes (does nothing)
		*/
	case 30:
		HL = 0;
		break;
		/*
		C = 31 (1Fh) : Get ADDR(Disk Parms)
		*/
	case 31:
		HL = DPBaddr;
		break;
		/*
		C = 32 (20h) : Get/Set user code
		*/
	case 32:
		if (LOW_REGISTER(DE) == 0xFF) {
			HL = userCode;
		} else {
			_SetUser(DE);
		}
		break;
		/*
		C = 33 (21h) : Read random
		*/
	case 33:
		HL = _ReadRand(DE);
		break;
		/*
		C = 34 (22h) : Write random
		*/
	case 34:
		HL = _WriteRand(DE);
		break;
		/*
		C = 35 (23h) : Compute file size
		*/
	case 35:
		HL = _GetFileSize(DE);
		break;
		/*
		C = 36 (24h) : Set random record
		*/
	case 36:
		HL = _SetRandom(DE);
		break;
		/*
		C = 37 (25h) : Reset drive
		*/
	case 37:
		break;
		/********** Function 38: Not supported by CP/M 2.2 **********/
		/********** Function 39: Not supported by CP/M 2.2 **********/
		/********** (todo) Function 40: Write random with zero fill **********/
		/*
		C = 40 (28h) : Write random with zero fill (we have no disk blocks, so just write random)
		*/
	case 40:
		HL = _WriteRand(DE);
		break;
// ##### added
// BDOS functions 39, 45, 47, 48, 98-103 are compatible with ZSDOS
		/*
		C = 39 (27h) : Get fixed disk vector
		*/
	case 39:
	  HL = 0xffff; // Disks A-P all are fixed
		break;
		/*
		C = 45 (2Dh) : Set error mode
		*/
	case 45:
	  // No action
		break;
		/*
		C = 47 (2Fh) : Get DMA address
		*/
	case 47:
	  HL = dmaAddr;
		break;
		/*
		C = 48 (30h) : Get DOS & version
		*/
	case 48:
	  HL = 0x5311; // ZSDOS 1.1
		break;
		/*
		C = 98 (62h) : Get time
		*/
	case 98:
		HL = _GetTime(DE);
		break;
		/*
		C = 99 (63h) : Set time
		*/
	case 99:
	  HL = _SetTime(DE);
		break;
		/*
		C = 100 (64h) : Get flags
		*/
	case 100:
		HL = 0;
		break;
		/*
		C = 101 (65h) : Set flags
		*/
	case 101:
		// No action
		break;
		/*
		C = 102 (66h) : Get file stamp
		*/
	case 102:
		HL = _GetFileStamp(DE, dmaAddr);
		break;
		/*
		C = 103 (67h) : Set file stamp
		*/
	case 103:
		HL =_SetFileStamp(DE, dmaAddr);
		break;
		/*
		C = 200 (C8h) : Write/Read EEPROM data
		D = 0xff -> write, D = 0x00 -> read
		E = data
		*/
	case 200:
		if (HIGH_REGISTER(DE)) {
			writeEEPROM(dmaAddr, LOW_REGISTER(DE));
		} else {
			HL = readEEPROM(dmaAddr);
		}
		break;
		/*
		C = 201 (C9h) : Set sound pitch
		DE = pitch [Hz]
		*/
	case 201:
		freq = DE;
		break;
		/*
		C = 202 (CAh) : Play sound
		DE = duration [ms]
		*/
	case 202:
		TimerFreeTone(BUZZERPIN, freq, DE);
		break;
// ##### end
		/*
		C = 220 (DCh) : PinMode
		*/
	case 220:
		pinMode(HIGH_REGISTER(DE), LOW_REGISTER(DE));
		break;
		/*
		C = 221 (DDh) : DigitalRead
		*/
	case 221:
		HL = digitalRead(HIGH_REGISTER(DE));
		break;
		/*
		C = 222 (DEh) : DigitalWrite
		*/
	case 222:
		digitalWrite(HIGH_REGISTER(DE), LOW_REGISTER(DE));
		break;
		/*
		C = 223 (DFh) : AnalogRead
		*/
	case 223:
		HL = analogRead(HIGH_REGISTER(DE));
		break;
		/*
		C = 224 (E0h) : AnalogWrite
		*/
	case 224:
		analogWrite(HIGH_REGISTER(DE), LOW_REGISTER(DE));
		break;
		/*
		C = 230 (E6h) : Set 8 bit masking
		*/
	case 230:
		mask8bit = LOW_REGISTER(DE);
		break;
		/*
		C = 249 (F9h) : MakeDisk
		Makes a disk directory if not existent.
		*/
	case 249:
		HL = _MakeDisk(DE);
		break;
		/*
		C = 250 (FAh) : HostOS
		Returns: A = 0x00 - Windows / 0x01 - Arduino / 0x02 - Posix / 0x03 - Dos / 0x04 - Teensy / 0x05 - ESP32 / 0x06 - STM32
		*/
	case 250:
		HL = 0x01;
		break;
		/*
		C = 251 (FBh) : Version
		Returns: A = 0xVv - Version in BCD representation: V.v
		*/
	case 251:
		HL = VersionBCD;
		break;
		/*
		C = 252 (FCh) : CCP version
		Returns: A = 0x00-0x04 = DRI|CCPZ|ZCPR2|ZCPR3|Z80CCP / 0xVv = Internal version in BCD: V.v
		*/
	case 252:
		HL = VersionCCP;
		break;
		/*
		C = 253 (FDh) : CCP address
		*/
	case 253:
		HL = CCPaddr;
		break;
		/*
		Unimplemented calls get listed
		*/
	default:
		break;
	}

	// CP/M BDOS does this before returning
	SET_HIGH_REGISTER(BC, HIGH_REGISTER(HL));
	SET_HIGH_REGISTER(AF, LOW_REGISTER(HL));

}

#endif
