#ifndef GLOBALS_H
#define GLOBALS_H

/* RunCPM version for the greeting header */
#define VERSION	"4.3"
#define VersionBCD 0x43

#define STR_HELPER(x) #x
#define STR(x) STR_HELPER(x)

/* Definition of the CCP memory information */
#define CCPname		"CCP-ZCP3." STR(TPASIZE) "K"
#define VersionCCP	0x03
#define BatchFCB  (CCPaddr + 0x63)    // Position of the $$$.SUB fcb on ZCPR33, variable!!!
#define CCPaddr		(BDOSpage-0x0800)   // ##### changed (standard 2 K CCP)
#define CCPHEAD		"\r\nRunCPM Version " VERSION " (CP/M 2.2 " STR(TPASIZE) "K)\r\n"
#define IObyte 0x0003
#define DUbyte 0x0004

#define BATCHA			// If this is defined, the $$$.SUB will be looked for on drive A:
//#define BATCH0		// If this is defined, the $$$.SUB will be looked for on user area 0
// The default behavior of DRI's CP/M 2.2 was to have $$$.SUB created on the current drive/user while looking for it
// on drive A: current user, which made it complicated to run SUBMITs when not logged to drive A: user 0

/* Some environment and type definitions */
typedef signed char     int8;
typedef signed short    int16;
typedef signed int      int32;
typedef unsigned char   uint8;
typedef unsigned short  uint16;
typedef unsigned int    uint32;

#define LOW_DIGIT(x)            ((x) & 0xf)
#define HIGH_DIGIT(x)           (((x) >> 4) & 0xf)
#define LOW_REGISTER(x)         ((x) & 0xff)
#define HIGH_REGISTER(x)        (((x) >> 8) & 0xff)
#define SET_LOW_REGISTER(x, v)  x = (((x) & 0xff00) | ((v) & 0xff))
#define SET_HIGH_REGISTER(x, v) x = (((x) & 0xff) | (((v) & 0xff) << 8))

#define WORD16(x)	((x) & 0xffff)

/* CP/M disk definitions */
#define BlkSZ 128	// CP/M block size
#define BlkEX 128	// Number of blocks on an extension
#define ExtSZ (BlkSZ * BlkEX)
#define BlkS2 4096	// Number of blocks on a S2 (module)
#define MaxEX 31	// Maximum value the EX field can take
#define MaxS2 15	// Maximum value the S2 (modules) field can take - Can be set to 63 to emulate CP/M Plus
#define MaxCR 128	// Maximum value the CR field can take
#define MaxRC 128	// Maximum value the RC field can take

/* CP/M memory definitions */
#define TPASIZE 60	      // Can be 60 for CP/M 2.2 compatibility or more, up to 64 for extra memory
#define MEMSIZE 64 * 1024	// RAM(plus ROM) needs to be 64K to avoid compatibility issues

static uint8 RAM[MEMSIZE];

#define _RamSysAddr(a)     &RAM[a]
#define _RamRead(a)        RAM[a]
#define _RamRead16(a)      ((RAM[(a & 0xffff) + 1] << 8) | RAM[a & 0xffff])
#define _RamWrite(a, v)    RAM[a] = v
#define _RamWrite16(a, v)  RAM[a] = (v) & 0xff; RAM[a + 1] = (v) >> 8
#define tohex(x)           ((x) < 10 ? (x) + 48 : (x) + 87)

// Size of the allocated pages (Minimum size = 1 page = 256 bytes)

// BIOS Pages (always on the top of memory)
#define BIOSpage     (MEMSIZE - 1280)   // ##### changed
#define BIOSjmppage  (BIOSpage - 256)

// BDOS Pages (depend on TPASIZE)
#define BDOSpage (TPASIZE * 1024) - 1024
#define DPBaddr (BIOSpage + 64)	// Address of the Disk Parameters Block (Hardcoded in BIOS)
#define SCBaddr 0xf980	// Address of the System Control Block ##### changed, see memory map
#define tmpFCB  0xf9c0	// Address of the temporary FCB        ##### changed, see memory map

/* Definition of global variables */
static uint8	filename[17];		// Current filename in host filesystem format
static uint8	newname[17];		// New filename in host filesystem format
static uint8	fcbname[13];		// Current filename in CP/M format
static uint8	pattern[13];		// File matching pattern in CP/M format
static uint16	dmaAddr = 0x0080;	// Current dmaAddr
static uint8	oDrive = 0;			// Old selected drive
static uint8	cDrive = 0;			// Currently selected drive
static uint8	userCode = 0;		// Current user code
static uint16	roVector = 0;
static uint16	loginVector = 0;
static uint8	allUsers = false;	// true when dr is '?' in BDOS search first
static uint8	allExtents = false; // true when ex is '?' in BDOS search first
static uint8	currFindUser = 0;	// user number of current directory in BDOS search first on all user numbers
static uint8	blockShift;			// disk allocation block shift
static uint8	blockMask;			// disk allocation block mask
static uint8	extentMask;			// disk extent mask
static uint16	firstBlockAfterDir;	// first allocation block after directory
static uint16	numAllocBlocks;		// # of allocation blocks on disk
static uint8	extentsPerDirEntry;	// # of logical (16K) extents in a directory entry
#define logicalExtentBytes (16*1024UL)
static uint16	physicalExtentBytes;	// # bytes described by 1 directory entry

/* Definition of externs to prevent precedence compilation errors */
extern void _Bdos();
extern void _Bios();
extern void _HostnameToFCB(uint16 fcbaddr, uint8* filename);
extern void _HostnameToFCBname(uint8* from, uint8* to);
extern void _mockupDirEntry();
extern uint8 match(uint8* fcbname, uint8* pattern);
extern void _puts(const char* str);

#endif
